#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pytest
import tests.resources
from blockchainetl_common.thread_local_proxy import ThreadLocalProxy
from iconetl.jobs.export_blocks_job import ExportBlocksJob
from iconetl.jobs.exporters.blocks_and_transactions_item_exporter import \
    blocks_and_transactions_item_exporter
from tests.iconetl.job.utils import get_web3_provider
from tests.utils import (compare_lines_ignore_order, read_file,
                         skip_if_slow_tests_disabled)

RESOURCE_GROUP = "test_export_blocks_job"


def read_resource(resource_group, file_name):
    return tests.resources.read_resource([RESOURCE_GROUP, resource_group], file_name)


@pytest.mark.parametrize(
    "start_block,end_block,batch_size,resource_group,web3_provider_type",
    [
        (10324748, 10324748, 1, "version_01a_block", "mock"),
        (12640760, 12640760, 1, "version_03_block", "mock"),
        (14473621, 14473621, 1, "version_04_block", "mock"),
        (14473622, 14473622, 1, "version_05_block", "mock"),
        skip_if_slow_tests_disabled(
            (10324748, 10324748, 1, "version_01a_block", "public_endpoint")
        ),
        skip_if_slow_tests_disabled(
            (12640760, 12640760, 1, "version_03_block", "public_endpoint")
        ),
        skip_if_slow_tests_disabled(
            (14473621, 14473621, 1, "version_04_block", "public_endpoint")
        ),
        skip_if_slow_tests_disabled(
            (14473622, 14473622, 1, "version_05_block", "public_endpoint")
        ),
    ],
)
def test_export_blocks_job(
    tmpdir, start_block, end_block, batch_size, resource_group, web3_provider_type
):
    blocks_output_file = str(tmpdir.join("actual_blocks.csv"))
    transactions_output_file = str(tmpdir.join("actual_transactions.csv"))

    job = ExportBlocksJob(
        start_block=start_block,
        end_block=end_block,
        batch_size=batch_size,
        batch_web3_provider=ThreadLocalProxy(
            lambda: get_web3_provider(
                web3_provider_type,
                lambda file: read_resource(resource_group, file),
                batch=True,
            )
        ),
        max_workers=5,
        item_exporter=blocks_and_transactions_item_exporter(
            blocks_output_file, transactions_output_file
        ),
        export_blocks=blocks_output_file is not None,
        export_transactions=transactions_output_file is not None,
    )
    job.run()

    compare_lines_ignore_order(
        read_resource(resource_group, "expected_blocks.csv"),
        read_file(blocks_output_file),
    )

    compare_lines_ignore_order(
        read_resource(resource_group, "expected_transactions.csv"),
        read_file(transactions_output_file),
    )



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from blockchainetl_common.jobs.exporters.composite_item_exporter import \
    CompositeItemExporter

BLOCK_FIELDS_TO_EXPORT = [
    "number",
    "hash",
    "parent_hash",
    "merkle_root_hash",
    "timestamp",
    "version",
    "peer_id",
    "signature",
    "next_leader",
]

TRANSACTION_FIELDS_TO_EXPORT = [
    "version",
    "from_address",
    "to_address",
    "value",
    "step_limit",
    "timestamp",
    "nid",
    "nonce",
    "hash",
    "transaction_index",
    "block_hash",
    "block_number",
    "fee",
    "signature",
    "data_type",
    "data",
]


def blocks_and_transactions_item_exporter(blocks_output=None, transactions_output=None):
    return CompositeItemExporter(
        filename_mapping={"block": blocks_output, "transaction": transactions_output},
        field_mapping={
            "block": BLOCK_FIELDS_TO_EXPORT,
            "transaction": TRANSACTION_FIELDS_TO_EXPORT,
        },
    )



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from iconetl.domain.transaction import IcxTransaction
from iconetl.utils import fix_tx_hash, hex_to_dec


class IcxTransactionMapper(object):
    def json_dict_to_transaction(
        self, json_dict, idx, block_hash, block_number, block_timestamp
    ):
        transaction = IcxTransaction()

        transaction.version = json_dict.get("version")
        transaction.from_address = json_dict.get("from")
        transaction.to_address = json_dict.get("to")
        transaction.value = hex_to_dec(json_dict.get("value", 0))
        transaction.step_limit = hex_to_dec(json_dict.get("stepLimit"))
        transaction.timestamp = json_dict.get("timestamp")
        transaction.nid = hex_to_dec(json_dict.get("nid"))
        transaction.nonce = hex_to_dec(json_dict.get("nonce"))
        if "txHash" in json_dict:
            transaction.hash = fix_tx_hash(json_dict.get("txHash"))

        if "tx_hash" in json_dict:
            transaction.hash = fix_tx_hash(json_dict.get("tx_hash"))

        if json_dict.get("txIndex"):
            transaction.transaction_index = hex_to_dec(json_dict.get("txIndex"))
        else:
            transaction.transaction_index = idx

        transaction.block_hash = block_hash
        transaction.block_number = block_number
        transaction.block_timestamp = block_timestamp
        transaction.fee = hex_to_dec(json_dict.get("fee"))
        transaction.signature = json_dict.get("signature")
        transaction.data_type = json_dict.get("dataType")
        transaction.data = json_dict.get("data")

        return transaction

    def transaction_to_dict(self, transaction):
        return {
            "type": "transaction",
            "version": transaction.version,
            "from_address": transaction.from_address,
            "to_address": transaction.to_address,
            "value": transaction.value,
            "step_limit": transaction.step_limit,
            "timestamp": transaction.timestamp,
            "block_timestamp": transaction.block_timestamp,
            "nid": transaction.nid,
            "nonce": transaction.nonce,
            "hash": transaction.hash,
            "transaction_index": transaction.transaction_index,
            "block_hash": transaction.block_hash,
            "block_number": transaction.block_number,
            "fee": transaction.fee,
            "signature": transaction.signature,
            "data_type": transaction.data_type,
            "data": transaction.data,
        }



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from datetime import datetime, timezone

from iconetl.service.graph_operations import (GraphOperations,
                                              OutOfBoundsError, Point)


class IcxService(object):
    def __init__(self, svc):
        graph = BlockTimestampGraph(svc)
        self._graph_operations = GraphOperations(graph)

    def get_block_range_for_date(self, date):
        start_datetime = datetime.combine(
            date, datetime.min.time().replace(tzinfo=timezone.utc)
        )
        end_datetime = datetime.combine(
            date, datetime.max.time().replace(tzinfo=timezone.utc)
        )
        return self.get_block_range_for_timestamps(
            start_datetime.timestamp(), end_datetime.timestamp()
        )

    def get_block_range_for_timestamps(self, start_timestamp, end_timestamp):
        start_timestamp = int(start_timestamp)
        end_timestamp = int(end_timestamp)
        if start_timestamp > end_timestamp:
            raise ValueError(
                "start_timestamp must be greater or equal to end_timestamp"
            )

        try:
            start_block_bounds = self._graph_operations.get_bounds_for_y_coordinate(
                start_timestamp
            )
        except OutOfBoundsError:
            start_block_bounds = (0, 0)

        try:
            end_block_bounds = self._graph_operations.get_bounds_for_y_coordinate(
                end_timestamp
            )
        except OutOfBoundsError as e:
            raise OutOfBoundsError(
                "The existing blocks do not completely cover the given time range"
            ) from e

        if (
            start_block_bounds == end_block_bounds
            and start_block_bounds[0] != start_block_bounds[1]
        ):
            raise ValueError("The given timestamp range does not cover any blocks")

        start_block = start_block_bounds[1]
        end_block = end_block_bounds[0]

        # The genesis block has timestamp 0 but we include it with the 1st block.
        if start_block == 1:
            start_block = 0

        return start_block, end_block


class BlockTimestampGraph(object):
    def __init__(self, svc):
        self._svc = svc

    def get_first_point(self):
        # Ignore the genesis block as its timestamp is 0
        return block_to_point(self._svc.get_block(1))

    def get_last_point(self):
        return block_to_point(self._svc.get_block("latest"))

    def get_point(self, x):
        return block_to_point(self._svc.get_block(x))


def block_to_point(block):
    return Point(block["height"], block["time_stamp"] / 1000000)



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os

from iconetl.providers.rpc import BatchHTTPProvider
from tests.iconetl.job.mock_batch_web3_provider import MockBatchWeb3Provider
from tests.iconetl.job.mock_web3_provider import MockWeb3Provider
from web3 import HTTPProvider


def get_web3_provider(provider_type, read_resource_lambda=None, batch=False):
    if provider_type == "mock":
        if read_resource_lambda is None:
            raise ValueError(
                "read_resource_lambda must not be None for provider type mock".format(
                    provider_type
                )
            )
        if batch:
            provider = MockBatchWeb3Provider(read_resource_lambda)
        else:
            provider = MockWeb3Provider(read_resource_lambda)
    elif provider_type == "public_endpoint":
        provider_url = os.environ.get(
            "PROVIDER_URL", "https://ctz.solidwallet.io/api/v3"
        )
        if batch:
            provider = BatchHTTPProvider(provider_url)
        else:
            provider = HTTPProvider(provider_url)
    else:
        raise ValueError("Provider type {} is unexpected".format(provider_type))
    return provider



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import csv
import sys


def set_max_field_size_limit():
    max_int = sys.maxsize
    decrement = True
    while decrement:
        # decrease the maxInt value by factor 10
        # as long as the OverflowError occurs.

        decrement = False
        try:
            csv.field_size_limit(max_int)
        except OverflowError:
            max_int = int(max_int / 10)
            decrement = True



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os


def read_resource(groups, file_name):
    current_file_dir = os.path.dirname(__file__)
    fixture_file_name = os.path.join(current_file_dir, *groups, file_name)

    if not os.path.exists(fixture_file_name):
        raise ValueError("File does not exist: " + fixture_file_name)

    with open(fixture_file_name, encoding="utf-8") as file_handle:
        return file_handle.read()



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from urllib.parse import urlparse

from iconetl.providers.rpc import BatchHTTPProvider
from web3 import HTTPProvider

DEFAULT_TIMEOUT = 60


def get_provider_from_uri(uri_string, timeout=DEFAULT_TIMEOUT, batch=False):
    uri = urlparse(uri_string)
    if uri.scheme == "http" or uri.scheme == "https":
        request_kwargs = {"timeout": timeout}
        if batch:
            return BatchHTTPProvider(uri_string, request_kwargs=request_kwargs)
        else:
            return HTTPProvider(uri_string, request_kwargs=request_kwargs)
    else:
        raise ValueError("Unknown uri scheme {}".format(uri_string))



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import csv

import click
from blockchainetl_common.file_utils import smart_open

from iconetl.csv_utils import set_max_field_size_limit


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "-i",
    "--input",
    default="-",
    show_default=True,
    type=str,
    help="The input file. If not specified stdin is used.",
)
@click.option(
    "-o",
    "--output",
    default="-",
    show_default=True,
    type=str,
    help="The output file. If not specified stdout is used.",
)
@click.option(
    "-c", "--column", required=True, type=str, help="The csv column name to extract."
)
def extract_csv_column(input, output, column):
    """Extracts column from given CSV file. Deprecated - use extract_field."""
    set_max_field_size_limit()

    with smart_open(input, "r") as input_file, smart_open(output, "w") as output_file:
        reader = csv.DictReader(input_file)
        for row in reader:
            if row[column]:
                output_file.write(row[column] + "\n")



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from iconetl.domain.receipt_log import IcxReceiptLog


class IcxReceiptLogMapper(object):
    def json_dict_to_receipt_log(
        self, json_dict, log_idx, tx_hash, tx_idx, blk_hash, blk_num
    ):
        receipt_log = IcxReceiptLog()

        receipt_log.log_index = log_idx
        receipt_log.transaction_hash = tx_hash
        receipt_log.transaction_index = tx_idx
        receipt_log.block_hash = blk_hash
        receipt_log.block_number = blk_num
        receipt_log.address = json_dict.get("scoreAddress")
        unsanitized_data = json_dict.get("data")
        sanitized_data = []
        for item in unsanitized_data:
            if item:
                sanitized_data.append(item.replace("\n", "").replace('"', "'"))
        receipt_log.data = sanitized_data
        receipt_log.indexed = json_dict.get("indexed")

        return receipt_log

    def receipt_log_to_dict(self, receipt_log):
        return {
            "type": "log",
            "log_index": receipt_log.log_index,
            "transaction_hash": receipt_log.transaction_hash,
            "transaction_index": receipt_log.transaction_index,
            "block_hash": receipt_log.block_hash,
            "block_number": receipt_log.block_number,
            "address": receipt_log.address,
            "data": receipt_log.data,
            "indexed": receipt_log.indexed,
        }

    def dict_to_receipt_log(self, dict):
        receipt_log = IcxReceiptLog()

        receipt_log.log_index = dict.get("log_index")
        receipt_log.transaction_hash = dict.get("transaction_hash")
        receipt_log.transaction_index = dict.get("transaction_index")
        receipt_log.block_hash = dict.get("block_hash")
        receipt_log.block_number = dict.get("block_number")
        receipt_log.address = dict.get("address")
        receipt_log.data = dict.get("data")
        receipt_log.indexed = dict.get("indexed")

        return receipt_log



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from iconetl.domain.receipt import IcxReceipt
from iconetl.mappers.receipt_log_mapper import IcxReceiptLogMapper
from iconetl.utils import hex_to_dec, to_normalized_address


class IcxReceiptMapper(object):
    def __init__(self, receipt_log_mapper=None):
        if receipt_log_mapper is None:
            self.receipt_log_mapper = IcxReceiptLogMapper()
        else:
            self.receipt_log_mapper = receipt_log_mapper

    def json_dict_to_receipt(self, json_dict):
        receipt = IcxReceipt()

        receipt.transaction_hash = json_dict.get("txHash")
        receipt.transaction_index = hex_to_dec(json_dict.get("txIndex"))
        receipt.block_hash = json_dict.get("blockHash")
        receipt.block_number = hex_to_dec(json_dict.get("blockHeight"))
        receipt.cumulative_step_used = hex_to_dec(json_dict.get("cumulativeStepUsed"))
        receipt.step_used = hex_to_dec(json_dict.get("stepUsed"))
        receipt.step_price = hex_to_dec(json_dict.get("stepPrice"))
        receipt.score_address = to_normalized_address(json_dict.get("scoreAddress"))
        receipt.status = hex_to_dec(json_dict.get("status"))

        if "eventLogs" in json_dict:
            receipt.logs = [
                self.receipt_log_mapper.json_dict_to_receipt_log(
                    log,
                    idx,
                    receipt.transaction_hash,
                    receipt.transaction_index,
                    receipt.block_hash,
                    receipt.block_number,
                )
                for idx, log in enumerate(json_dict["eventLogs"])
            ]

        return receipt

    def receipt_to_dict(self, receipt):
        return {
            "type": "receipt",
            "transaction_hash": receipt.transaction_hash,
            "transaction_index": receipt.transaction_index,
            "block_hash": receipt.block_hash,
            "block_number": receipt.block_number,
            "cumulative_step_used": receipt.cumulative_step_used,
            "step_used": receipt.step_used,
            "step_price": receipt.step_price,
            "score_address": receipt.score_address,
            "status": receipt.status,
        }



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import itertools

from iconetl.misc.retriable_value_error import RetriableValueError


def hex_to_dec(hex_string):
    if hex_string is None:
        return None
    if isinstance(hex_string, int):
        return hex_string
    try:
        return int(hex_string, 16)
    except ValueError:
        print("Not a hex string %s" % hex_string)
        return hex_string


def to_normalized_address(address):
    if address is None or not isinstance(address, str):
        return address
    return address.lower()


def fix_tx_hash(hash):
    if hash[0:2] != "0x":
        return "0x" + hash
    return hash


def validate_range(range_start_incl, range_end_incl):
    if range_start_incl < 0 or range_end_incl < 0:
        raise ValueError("range_start and range_end must be greater or equal to 0")

    if range_end_incl < range_start_incl:
        raise ValueError("range_end must be greater or equal to range_start")


def rpc_response_batch_to_results(response):
    for response_item in response:
        yield rpc_response_to_result(response_item)


def rpc_response_to_result(response):
    result = response.get("result")
    if result is None:
        error_message = "result is None in response {}.".format(response)
        if response.get("error") is None:
            error_message = error_message + " Make sure node is synced."
            raise RetriableValueError(error_message)
        elif response.get("error") is not None and is_retriable_error(
            response.get("error").get("code")
        ):
            raise RetriableValueError(error_message)
        raise ValueError(error_message)
    return result


def is_retriable_error(error_code):
    if error_code is None:
        return False

    if not isinstance(error_code, int):
        return False

    if error_code == -32603 or (-32000 >= error_code >= -32099):
        return True

    return False


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import click
from blockchainetl_common.file_utils import smart_open
from blockchainetl_common.logging_utils import logging_basic_config
from blockchainetl_common.thread_local_proxy import ThreadLocalProxy
from iconetl.jobs.export_receipts_job import ExportReceiptsJob
from iconetl.jobs.exporters.receipts_and_logs_item_exporter import \
    receipts_and_logs_item_exporter
from iconetl.providers.auto import get_provider_from_uri

logging_basic_config()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "-b",
    "--batch-size",
    default=100,
    show_default=True,
    type=int,
    help="The number of receipts to export at a time.",
)
@click.option(
    "-t",
    "--transaction-hashes",
    required=True,
    type=str,
    help="The file containing transaction hashes, one per line.",
)
@click.option(
    "-p",
    "--provider-uri",
    default="https://ctz.solidwallet.io/api/v3",
    show_default=True,
    type=str,
    help="The URI of the node endpoint",
)
@click.option(
    "-w",
    "--max-workers",
    default=5,
    show_default=True,
    type=int,
    help="The maximum number of workers.",
)
@click.option(
    "--receipts-output",
    default=None,
    show_default=True,
    type=str,
    help='The output file for receipts. If not provided receipts will not be exported. Use "-" for stdout',
)
@click.option(
    "--logs-output",
    default=None,
    show_default=True,
    type=str,
    help="The output file for receipt logs. "
    'aIf not provided receipt logs will not be exported. Use "-" for stdout',
)
def export_receipts_and_logs(
    batch_size,
    transaction_hashes,
    provider_uri,
    max_workers,
    receipts_output,
    logs_output,
):
    """Exports receipts and logs."""
    with smart_open(transaction_hashes, "r") as transaction_hashes_file:
        job = ExportReceiptsJob(
            transaction_hashes_iterable=(
                transaction_hash.strip() for transaction_hash in transaction_hashes_file
            ),
            batch_size=batch_size,
            batch_web3_provider=ThreadLocalProxy(
                lambda: get_provider_from_uri(provider_uri, batch=True)
            ),
            max_workers=max_workers,
            item_exporter=receipts_and_logs_item_exporter(receipts_output, logs_output),
            export_receipts=receipts_output is not None,
            export_logs=logs_output is not None,
        )

        job.run()



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import click
from blockchainetl_common.file_utils import smart_open
from blockchainetl_common.logging_utils import logging_basic_config
from iconetl.providers.auto import get_provider_from_uri
from iconetl.service.icx_service import IcxService
from iconsdk.icon_service import IconService
from iconsdk.providers.http_provider import HTTPProvider

logging_basic_config()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "-p",
    "--provider-uri",
    default="https://ctz.solidwallet.io/api/v3",
    show_default=True,
    type=str,
    help="The URI of the node endpoint",
)
@click.option(
    "-s",
    "--start-timestamp",
    required=True,
    type=int,
    help="Start unix timestamp, in seconds.",
)
@click.option(
    "-e",
    "--end-timestamp",
    required=True,
    type=int,
    help="End unix timestamp, in seconds.",
)
@click.option(
    "-o",
    "--output",
    default="-",
    type=str,
    help="The output file. If not specified stdout is used.",
)
def get_block_range_for_timestamps(
    provider_uri, start_timestamp, end_timestamp, output
):
    """Outputs start and end blocks for given timestamps."""
    provider = get_provider_from_uri(provider_uri)
    svc = IconService(HTTPProvider(provider))
    icx_service = IcxService(svc)

    start_block, end_block = icx_service.get_block_range_for_timestamps(
        start_timestamp, end_timestamp
    )

    with smart_open(output, "w") as output_file:
        output_file.write("{},{}\n".format(start_block, end_block))



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import click

from iconetl.cli.export_blocks_and_transactions import export_blocks_and_transactions
from iconetl.cli.export_receipts_and_logs import export_receipts_and_logs
from iconetl.cli.extract_csv_column import extract_csv_column
from iconetl.cli.get_block_range_for_date import get_block_range_for_date
from iconetl.cli.get_block_range_for_timestamps import get_block_range_for_timestamps


@click.group()
@click.version_option(version="0.1.1")
@click.pass_context
def cli(ctx):
    pass


cli.add_command(export_receipts_and_logs, "export_receipts_and_logs")
cli.add_command(export_blocks_and_transactions, "export_blocks_and_transactions")

cli.add_command(get_block_range_for_date, "get_block_range_for_date")
cli.add_command(get_block_range_for_timestamps, "get_block_range_for_timestamps")
cli.add_command(extract_csv_column, "extract_csv_column")



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import json

from blockchainetl_common.executors.batch_work_executor import \
    BatchWorkExecutor
from blockchainetl_common.jobs.base_job import BaseJob
from iconetl.json_rpc_requests import generate_get_block_by_number_json_rpc
from iconetl.mappers.block_mapper import IcxBlockMapper
from iconetl.mappers.transaction_mapper import IcxTransactionMapper
from iconetl.utils import rpc_response_batch_to_results, validate_range


class ExportBlocksJob(BaseJob):
    def __init__(
        self,
        start_block,
        end_block,
        batch_size,
        batch_web3_provider,
        max_workers,
        item_exporter,
        export_blocks=True,
        export_transactions=True,
    ):
        validate_range(start_block, end_block)
        self.start_block = start_block
        self.end_block = end_block

        self.batch_web3_provider = batch_web3_provider

        self.batch_work_executor = BatchWorkExecutor(batch_size, max_workers)
        self.item_exporter = item_exporter

        self.export_blocks = export_blocks
        self.export_transactions = export_transactions
        if not self.export_blocks and not self.export_transactions:
            raise ValueError(
                "At least one of export_blocks or export_transactions must be True"
            )

        self.block_mapper = IcxBlockMapper()
        self.transaction_mapper = IcxTransactionMapper()

    def _start(self):
        self.item_exporter.open()

    def _export(self):
        self.batch_work_executor.execute(
            range(self.start_block, self.end_block + 1),
            self._export_batch,
            total_items=self.end_block - self.start_block + 1,
        )

    def _export_batch(self, block_number_batch):
        blocks_rpc = list(generate_get_block_by_number_json_rpc(block_number_batch))
        response = self.batch_web3_provider.make_batch_request(json.dumps(blocks_rpc))
        results = rpc_response_batch_to_results(response)
        blocks = [self.block_mapper.json_dict_to_block(result) for result in results]

        for block in blocks:
            self._export_block(block)

    def _export_block(self, block):
        if self.export_blocks:
            self.item_exporter.export_item(self.block_mapper.block_to_dict(block))
        if self.export_transactions:
            for tx in block.transactions:
                self.item_exporter.export_item(
                    self.transaction_mapper.transaction_to_dict(tx)
                )

    def _end(self):
        self.batch_work_executor.shutdown()
        self.item_exporter.close()



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pytest
import tests.resources
from blockchainetl_common.thread_local_proxy import ThreadLocalProxy
from iconetl.jobs.export_receipts_job import ExportReceiptsJob
from iconetl.jobs.exporters.receipts_and_logs_item_exporter import \
    receipts_and_logs_item_exporter
from tests.iconetl.job.utils import get_web3_provider
from tests.utils import (compare_lines_ignore_order, read_file,
                         skip_if_slow_tests_disabled)

RESOURCE_GROUP = "test_export_receipts_job"


def read_resource(resource_group, file_name):
    return tests.resources.read_resource([RESOURCE_GROUP, resource_group], file_name)


DEFAULT_TX_HASHES = [
    "0x3680f3262fed98a4bc169c11b6ac66778da43c889d47a673d252362138715da9",
    "0xaec1f999ec61c80e23701229a33ec3b471c073e721d0a4096f37f2e1ea6bfa19",
    "0x3ce26161a89e0747b87df969ce76af1c275237f0e95a7415cca8e578ed7372d7",
    "0xcd159d87e580c51a770017a813dd451bed6cbe587ff6c84bc43a95efe9ea7c61",
    "0xe87509d050cd172b1ca8129b3ba2028c5ff84b53684f56401d01d09934777039",
    "0x341884ab68c291df94ca9ce8e4948b06ed80652519f0a228b30ec7bda64d001c",
]

HTML_TX_HASHES = [
    "0x16dbc932b601821b08450ad6f228a6a8e1bfd9cf5a361f0bf42ccf4b0b29be7b",
    "0x1a94bf51895fc4112ccd17f17a382fbe2cc5c0f787595256c98e0669b572c52e",
]


@pytest.mark.parametrize(
    "batch_size,transaction_hashes,output_format,resource_group,web3_provider_type",
    [
        (1, DEFAULT_TX_HASHES, "csv", "receipts_with_logs", "mock"),
        (2, DEFAULT_TX_HASHES, "csv", "receipts_with_logs", "mock"),
        (2, DEFAULT_TX_HASHES, "json", "receipts_with_logs", "mock"),
        skip_if_slow_tests_disabled(
            (1, DEFAULT_TX_HASHES, "csv", "receipts_with_logs", "public_endpoint")
        ),
        skip_if_slow_tests_disabled(
            (2, DEFAULT_TX_HASHES, "json", "receipts_with_logs", "public_endpoint")
        ),
        (2, HTML_TX_HASHES, "csv", "html_sanitize", "mock"),
        (2, HTML_TX_HASHES, "json", "html_sanitize", "mock"),
        skip_if_slow_tests_disabled(
            (2, HTML_TX_HASHES, "csv", "html_sanitize", "mock")
        ),
    ],
)
def test_export_receipts_job(
    tmpdir,
    batch_size,
    transaction_hashes,
    output_format,
    resource_group,
    web3_provider_type,
):
    receipts_output_file = str(tmpdir.join("actual_receipts." + output_format))
    logs_output_file = str(tmpdir.join("actual_logs." + output_format))

    job = ExportReceiptsJob(
        transaction_hashes_iterable=transaction_hashes,
        batch_size=batch_size,
        batch_web3_provider=ThreadLocalProxy(
            lambda: get_web3_provider(
                web3_provider_type,
                lambda file: read_resource(resource_group, file),
                batch=True,
            )
        ),
        max_workers=5,
        item_exporter=receipts_and_logs_item_exporter(
            receipts_output_file, logs_output_file
        ),
        export_receipts=receipts_output_file is not None,
        export_logs=logs_output_file is not None,
    )
    job.run()

    compare_lines_ignore_order(
        read_resource(resource_group, "expected_receipts." + output_format),
        read_file(receipts_output_file),
    )

    compare_lines_ignore_order(
        read_resource(resource_group, "expected_logs." + output_format),
        read_file(logs_output_file),
    )



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pytest
from dateutil.parser import parse
from iconetl.service.graph_operations import OutOfBoundsError
from iconetl.service.icx_service import IcxService
from iconsdk.icon_service import IconService
from iconsdk.providers.http_provider import HTTPProvider
from tests.utils import skip_if_slow_tests_disabled


@pytest.mark.parametrize(
    "date,expected_start_block,expected_end_block",
    [
        skip_if_slow_tests_disabled(["2018-02-07", 2, 18]),
        skip_if_slow_tests_disabled(["2018-12-31", 155728, 156584]),
        skip_if_slow_tests_disabled(["2019-01-01", 156585, 157566]),
        skip_if_slow_tests_disabled(["2019-12-31", 12975470, 13018226]),
        skip_if_slow_tests_disabled(["2020-01-01", 13018228, 13061082]),
        skip_if_slow_tests_disabled(["2020-01-02", 13061084, 13103934]),
    ],
)
def test_get_block_range_for_date(date, expected_start_block, expected_end_block):
    icx_service = get_new_icx_service()
    parsed_date = parse(date)
    blocks = icx_service.get_block_range_for_date(parsed_date)
    assert blocks == (expected_start_block, expected_end_block)


@pytest.mark.parametrize(
    "date",
    [
        skip_if_slow_tests_disabled(["2015-07-29"]),
        skip_if_slow_tests_disabled(["2030-01-01"]),
    ],
)
def test_get_block_range_for_date_fail(date):
    icx_service = get_new_icx_service()
    parsed_date = parse(date)
    with pytest.raises(OutOfBoundsError):
        icx_service.get_block_range_for_date(parsed_date)


@pytest.mark.parametrize(
    "start_timestamp,end_timestamp,expected_start_block,expected_end_block",
    [skip_if_slow_tests_disabled([1517999935, 1517999936, 16, 16]),],
)
def test_get_block_range_for_timestamps(
    start_timestamp, end_timestamp, expected_start_block, expected_end_block
):
    icx_service = get_new_icx_service()
    blocks = icx_service.get_block_range_for_timestamps(start_timestamp, end_timestamp)
    assert blocks == (expected_start_block, expected_end_block)


@pytest.mark.parametrize(
    "start_timestamp,end_timestamp",
    [skip_if_slow_tests_disabled([1517999934, 1517999935])],
)
def test_get_block_range_for_timestamps_fail(start_timestamp, end_timestamp):
    icx_service = get_new_icx_service()
    with pytest.raises(ValueError):
        icx_service.get_block_range_for_timestamps(start_timestamp, end_timestamp)


def get_new_icx_service():
    svc = IconService(HTTPProvider("https://ctz.solidwallet.io/api/v3"))
    return IcxService(svc)



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
from datetime import datetime, timedelta

import click
from blockchainetl_common.logging_utils import logging_basic_config
from iconsdk.icon_service import IconService
from iconsdk.providers.http_provider import HTTPProvider

from iconetl.jobs.export_all_common import export_all_common
from iconetl.service.icx_service import IcxService

logging_basic_config()


def is_date_range(start, end):
    """Checks for YYYY-MM-DD date format."""
    return bool(
        re.match("^2[0-9]{3}-[0-9]{2}-[0-9]{2}$", start)
        and re.match("^2[0-9]{3}-[0-9]{2}-[0-9]{2}$", end)
    )


def is_unix_time_range(start, end):
    """Checks for Unix timestamp format."""
    return bool(
        re.match("^[0-9]{10}$|^[0-9]{13}$", start)
        and re.match("^[0-9]{10}$|^[0-9]{13}$", end)
    )


def is_block_range(start, end):
    """Checks for a valid block number."""
    return (
        start.isdigit()
        and 0 <= int(start) <= 99999999
        and end.isdigit()
        and 0 <= int(end) <= 99999999
    )


def get_partitions(start, end, partition_batch_size, provider_uri):
    """Yield partitions based on input data type."""
    if is_date_range(start, end) or is_unix_time_range(start, end):
        if is_date_range(start, end):
            start_date = datetime.strptime(start, "%Y-%m-%d").date()
            end_date = datetime.strptime(end, "%Y-%m-%d").date()

        elif is_unix_time_range(start, end):
            if len(start) == 10 and len(end) == 10:
                start_date = datetime.utcfromtimestamp(int(start)).date()
                end_date = datetime.utcfromtimestamp(int(end)).date()

            elif len(start) == 13 and len(end) == 13:
                start_date = datetime.utcfromtimestamp(int(start) / 1e3).date()
                end_date = datetime.utcfromtimestamp(int(end) / 1e3).date()

        day = timedelta(days=1)

        svc = IconService(HTTPProvider(provider_uri))
        icx_service = IcxService(svc)

        while start_date <= end_date:
            batch_start_block, batch_end_block = icx_service.get_block_range_for_date(
                start_date
            )
            partition_dir = "/date={start_date!s}/".format(start_date=start_date)
            yield batch_start_block, batch_end_block, partition_dir
            start_date += day

    elif is_block_range(start, end):
        start_block = int(start)
        end_block = int(end)

        for batch_start_block in range(
            start_block, end_block + 1, partition_batch_size
        ):
            batch_end_block = batch_start_block + partition_batch_size - 1
            if batch_end_block > end_block:
                batch_end_block = end_block

            padded_batch_start_block = str(batch_start_block).zfill(8)
            padded_batch_end_block = str(batch_end_block).zfill(8)
            partition_dir = "/start_block={padded_batch_start_block}/end_block={padded_batch_end_block}".format(
                padded_batch_start_block=padded_batch_start_block,
                padded_batch_end_block=padded_batch_end_block,
            )
            yield batch_start_block, batch_end_block, partition_dir

    else:
        raise ValueError(
            "start and end must be either block numbers or ISO dates or Unix times"
        )


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "-s", "--start", required=True, type=str, help="Start block/ISO date/Unix time"
)
@click.option(
    "-e", "--end", required=True, type=str, help="End block/ISO date/Unix time"
)
@click.option(
    "-b",
    "--partition-batch-size",
    default=10000,
    show_default=True,
    type=int,
    help="The number of blocks to export in partition.",
)
@click.option(
    "-p",
    "--provider-uri",
    default="https://ctz.solidwallet.io/api/v3",
    show_default=True,
    type=str,
    help="The URI of the node endpoint",
)
@click.option(
    "-o",
    "--output-dir",
    default="output",
    show_default=True,
    type=str,
    help="Output directory, partitioned in Hive style.",
)
@click.option(
    "-w",
    "--max-workers",
    default=5,
    show_default=True,
    type=int,
    help="The maximum number of workers.",
)
@click.option(
    "-B",
    "--export-batch-size",
    default=100,
    show_default=True,
    type=int,
    help="The number of requests in JSON RPC batches.",
)
def export_all(
    start,
    end,
    partition_batch_size,
    provider_uri,
    output_dir,
    max_workers,
    export_batch_size,
):
    """Exports all data for a range of blocks."""
    export_all_common(
        get_partitions(start, end, partition_batch_size, provider_uri),
        output_dir,
        provider_uri,
        max_workers,
        export_batch_size,
    )



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import click
from blockchainetl_common.logging_utils import logging_basic_config
from blockchainetl_common.thread_local_proxy import ThreadLocalProxy
from iconetl.jobs.export_blocks_job import ExportBlocksJob
from iconetl.jobs.exporters.blocks_and_transactions_item_exporter import \
    blocks_and_transactions_item_exporter
from iconetl.providers.auto import get_provider_from_uri

logging_basic_config()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option(
    "-s", "--start-block", default=0, show_default=True, type=int, help="Start block",
)
@click.option(
    "-e", "--end-block", required=True, type=int, help="End block",
)
@click.option(
    "-b",
    "--batch-size",
    default=100,
    show_default=True,
    type=int,
    help="The number of blocks to export at a time.",
)
@click.option(
    "-p",
    "--provider-uri",
    default="https://ctz.solidwallet.io/api/v3",
    show_default=True,
    type=str,
    help="The URI of the node endpoint",
)
@click.option(
    "-w",
    "--max-workers",
    default=5,
    show_default=True,
    type=int,
    help="The maximum number of workers.",
)
@click.option(
    "--blocks-output",
    default=None,
    show_default=True,
    type=str,
    help='The output file for blocks. If not provided blocks will not be exported. Use "-" for stdout',
)
@click.option(
    "--transactions-output",
    default=None,
    show_default=True,
    type=str,
    help="The output file for transactions. "
    'If not provided transactions will not be exported. Use "-" for stdout',
)
def export_blocks_and_transactions(
    start_block,
    end_block,
    batch_size,
    provider_uri,
    max_workers,
    blocks_output,
    transactions_output,
):
    """Exports blocks and transactions."""
    if blocks_output is None and transactions_output is None:
        raise ValueError(
            "Either --blocks-output or --transactions-output options must be provided"
        )

    job = ExportBlocksJob(
        start_block=start_block,
        end_block=end_block,
        batch_size=batch_size,
        batch_web3_provider=ThreadLocalProxy(
            lambda: get_provider_from_uri(provider_uri, batch=True)
        ),
        max_workers=max_workers,
        item_exporter=blocks_and_transactions_item_exporter(
            blocks_output, transactions_output
        ),
        export_blocks=blocks_output is not None,
        export_transactions=transactions_output is not None,
    )
    job.run()



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import json

from blockchainetl_common.executors.batch_work_executor import \
    BatchWorkExecutor
from blockchainetl_common.jobs.base_job import BaseJob
from iconetl.json_rpc_requests import generate_get_receipt_json_rpc
from iconetl.mappers.receipt_log_mapper import IcxReceiptLogMapper
from iconetl.mappers.receipt_mapper import IcxReceiptMapper
from iconetl.utils import rpc_response_batch_to_results


# Exports receipts and logs
class ExportReceiptsJob(BaseJob):
    def __init__(
        self,
        transaction_hashes_iterable,
        batch_size,
        batch_web3_provider,
        max_workers,
        item_exporter,
        export_receipts=True,
        export_logs=True,
    ):
        self.batch_web3_provider = batch_web3_provider
        self.transaction_hashes_iterable = transaction_hashes_iterable

        self.batch_work_executor = BatchWorkExecutor(batch_size, max_workers)
        self.item_exporter = item_exporter

        self.export_receipts = export_receipts
        self.export_logs = export_logs
        if not self.export_receipts and not self.export_logs:
            raise ValueError(
                "At least one of export_receipts or export_logs must be True"
            )

        self.receipt_mapper = IcxReceiptMapper()
        self.receipt_log_mapper = IcxReceiptLogMapper()

    def _start(self):
        self.item_exporter.open()

    def _export(self):
        self.batch_work_executor.execute(
            self.transaction_hashes_iterable, self._export_receipts
        )

    def _export_receipts(self, transaction_hashes):
        receipts_rpc = list(generate_get_receipt_json_rpc(transaction_hashes))
        response = self.batch_web3_provider.make_batch_request(json.dumps(receipts_rpc))
        results = rpc_response_batch_to_results(response)
        receipts = [
            self.receipt_mapper.json_dict_to_receipt(result) for result in results
        ]
        for receipt in receipts:
            self._export_receipt(receipt)

    def _export_receipt(self, receipt):
        if self.export_receipts:
            self.item_exporter.export_item(self.receipt_mapper.receipt_to_dict(receipt))
        if self.export_logs:
            for log in receipt.logs:
                self.item_exporter.export_item(
                    self.receipt_log_mapper.receipt_log_to_dict(log)
                )

    def _end(self):
        self.batch_work_executor.shutdown()
        self.item_exporter.close()



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


def generate_get_block_by_number_json_rpc(block_numbers):
    for idx, block_number in enumerate(block_numbers):
        yield generate_json_rpc(
            method="icx_getBlockByHeight",
            params={"height": hex(block_number)},
            request_id=idx,
        )


def generate_get_receipt_json_rpc(transaction_hashes):
    for idx, transaction_hash in enumerate(transaction_hashes):
        yield generate_json_rpc(
            method="icx_getTransactionResult",
            params={"txHash": transaction_hash},
            request_id=idx,
        )


def generate_json_rpc(method, params, request_id=1):
    return {
        "jsonrpc": "2.0",
        "method": method,
        "params": params,
        "id": request_id,
    }



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from iconetl.domain.block import IcxBlock
from iconetl.mappers.transaction_mapper import IcxTransactionMapper
from iconetl.utils import hex_to_dec


class IcxBlockMapper(object):
    def __init__(self, transaction_mapper=None):
        if transaction_mapper is None:
            self.transaction_mapper = IcxTransactionMapper()
        else:
            self.transaction_mapper = transaction_mapper

    def json_dict_to_block(self, json_dict):
        block = IcxBlock()

        block.number = json_dict.get("height")
        block.hash = json_dict.get("block_hash")
        block.parent_hash = json_dict.get("prev_block_hash")
        block.merkle_root_hash = json_dict.get("merkle_tree_root_hash")
        block.timestamp = json_dict.get("time_stamp")
        block.version = json_dict.get("version")
        block.peer_id = json_dict.get("peer_id")
        block.signature = json_dict.get("signature")
        block.next_leader = json_dict.get("next_leader")

        if "confirmed_transaction_list" in json_dict:
            block.transactions = [
                self.transaction_mapper.json_dict_to_transaction(
                    tx, idx, block.hash, block.number, block.timestamp
                )
                for idx, tx in enumerate(json_dict["confirmed_transaction_list"])
                if isinstance(tx, dict)
            ]

        return block

    def block_to_dict(self, block):
        return {
            "type": "block",
            "number": block.number,
            "hash": block.hash,
            "parent_hash": block.parent_hash,
            "merkle_root_hash": block.merkle_root_hash,
            "timestamp": block.timestamp,
            "version": block.version,
            "transactions": block.transactions,
            "peer_id": block.peer_id,
            "signature": block.signature,
            "next_leader": block.next_leader,
        }



import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


long_description = read("README.md") if os.path.isfile("README.md") else ""

setup(
    name="icon-etl",
    version="0.1.1",
    packages=find_packages(exclude=["schemas", "tests"]),
    url="https://github.com/insight-icon/icon-etl",
    author="Richard Mah",
    author_email="richard@richardmah.com",
    description="Tools for exporting ICON blockchain data to CSV or JSON",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6.0,<3.8.0",
    install_requires=[
        "blockchain-etl-common==1.3.0",
        "requests==2.20.0",
        "python-dateutil==2.7.0",
        "click==7.0",
        "iconsdk >= 1.3.2",
        "web3 >= 5.5.0",
    ],
    extras_require={
        "streaming": ["timeout-decorator==0.4.1", "sqlalchemy==1.3.13",],
        "dev": ["pytest~=4.3.0"],
    },
    project_urls={
        "Bug Reports": "https://github.com/insight-icon/icon-etl/issues",
        "Source": "https://github.com/insight-icon/icon-etl",
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": ["iconetl=iconetl.cli:cli",],},
)



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import csv
import logging
import os
import shutil
from time import time

from blockchainetl_common.file_utils import smart_open
from blockchainetl_common.thread_local_proxy import ThreadLocalProxy

from iconetl.csv_utils import set_max_field_size_limit
from iconetl.jobs.export_blocks_job import ExportBlocksJob
from iconetl.jobs.export_receipts_job import ExportReceiptsJob
from iconetl.jobs.exporters.blocks_and_transactions_item_exporter import (
    blocks_and_transactions_item_exporter,
)
from iconetl.jobs.exporters.receipts_and_logs_item_exporter import (
    receipts_and_logs_item_exporter,
)
from iconetl.providers.auto import get_provider_from_uri

logger = logging.getLogger("export_all")


def extract_csv_column_unique(input, output, column):
    set_max_field_size_limit()

    with smart_open(input, "r") as input_file, smart_open(output, "w") as output_file:
        reader = csv.DictReader(input_file)
        seen = set()
        for row in reader:
            if row[column] in seen:
                continue
            seen.add(row[column])
            output_file.write(row[column] + "\n")


def export_all_common(partitions, output_dir, provider_uri, max_workers, batch_size):
    for batch_start_block, batch_end_block, partition_dir in partitions:

        start_time = time()

        padded_batch_start_block = str(batch_start_block).zfill(8)
        padded_batch_end_block = str(batch_end_block).zfill(8)
        block_range = "{padded_batch_start_block}-{padded_batch_end_block}".format(
            padded_batch_start_block=padded_batch_start_block,
            padded_batch_end_block=padded_batch_end_block,
        )
        file_name_suffix = "{padded_batch_start_block}_{padded_batch_end_block}".format(
            padded_batch_start_block=padded_batch_start_block,
            padded_batch_end_block=padded_batch_end_block,
        )

        blocks_output_dir = "{output_dir}/blocks{partition_dir}".format(
            output_dir=output_dir, partition_dir=partition_dir,
        )
        os.makedirs(os.path.dirname(blocks_output_dir), exist_ok=True)

        transactions_output_dir = "{output_dir}/transactions{partition_dir}".format(
            output_dir=output_dir, partition_dir=partition_dir,
        )
        os.makedirs(os.path.dirname(transactions_output_dir), exist_ok=True)

        blocks_file = "{blocks_output_dir}/blocks_{file_name_suffix}.csv".format(
            blocks_output_dir=blocks_output_dir, file_name_suffix=file_name_suffix,
        )
        transactions_file = "{transactions_output_dir}/transactions_{file_name_suffix}.csv".format(
            transactions_output_dir=transactions_output_dir,
            file_name_suffix=file_name_suffix,
        )
        logger.info(
            "Exporting blocks {block_range} to {blocks_file}".format(
                block_range=block_range, blocks_file=blocks_file,
            )
        )
        logger.info(
            "Exporting transactions from blocks {block_range} to {transactions_file}".format(
                block_range=block_range, transactions_file=transactions_file,
            )
        )

        job = ExportBlocksJob(
            start_block=batch_start_block,
            end_block=batch_end_block,
            batch_size=batch_size,
            batch_web3_provider=ThreadLocalProxy(
                lambda: get_provider_from_uri(provider_uri, batch=True)
            ),
            max_workers=max_workers,
            item_exporter=blocks_and_transactions_item_exporter(
                blocks_file, transactions_file
            ),
            export_blocks=blocks_file is not None,
            export_transactions=transactions_file is not None,
        )
        job.run()

        cache_output_dir = "{output_dir}/.tmp{partition_dir}".format(
            output_dir=output_dir, partition_dir=partition_dir,
        )
        os.makedirs(os.path.dirname(cache_output_dir), exist_ok=True)

        transaction_hashes_file = "{cache_output_dir}/transaction_hashes_{file_name_suffix}.csv".format(
            cache_output_dir=cache_output_dir, file_name_suffix=file_name_suffix,
        )
        logger.info(
            "Extracting hash column from transaction file {transactions_file}".format(
                transactions_file=transactions_file,
            )
        )
        extract_csv_column_unique(transactions_file, transaction_hashes_file, "hash")

        receipts_output_dir = "{output_dir}/receipts{partition_dir}".format(
            output_dir=output_dir, partition_dir=partition_dir,
        )
        os.makedirs(os.path.dirname(receipts_output_dir), exist_ok=True)

        logs_output_dir = "{output_dir}/logs{partition_dir}".format(
            output_dir=output_dir, partition_dir=partition_dir,
        )
        os.makedirs(os.path.dirname(logs_output_dir), exist_ok=True)

        receipts_file = "{receipts_output_dir}/receipts_{file_name_suffix}.csv".format(
            receipts_output_dir=receipts_output_dir, file_name_suffix=file_name_suffix,
        )
        logs_file = "{logs_output_dir}/logs_{file_name_suffix}.csv".format(
            logs_output_dir=logs_output_dir, file_name_suffix=file_name_suffix,
        )
        logger.info(
            "Exporting receipts and logs from blocks {block_range} to {receipts_file} and {logs_file}".format(
                block_range=block_range,
                receipts_file=receipts_file,
                logs_file=logs_file,
            )
        )

        with smart_open(transaction_hashes_file, "r") as transaction_hashes:
            job = ExportReceiptsJob(
                transaction_hashes_iterable=(
                    transaction_hash.strip() for transaction_hash in transaction_hashes
                ),
                batch_size=batch_size,
                batch_web3_provider=ThreadLocalProxy(
                    lambda: get_provider_from_uri(provider_uri, batch=True)
                ),
                max_workers=max_workers,
                item_exporter=receipts_and_logs_item_exporter(receipts_file, logs_file),
                export_receipts=receipts_file is not None,
                export_logs=logs_file is not None,
            )
            job.run()

        shutil.rmtree(os.path.dirname(cache_output_dir))
        end_time = time()
        time_diff = round(end_time - start_time, 5)
        logger.info(
            "Exporting blocks {block_range} took {time_diff} seconds".format(
                block_range=block_range, time_diff=time_diff,
            )
        )



#  MIT License
#
#  Copyright (c) 2020 Richard Mah (richard@richardmah.com) & Insight Infrastructure
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
#  the Software, and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
#  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
#  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from web3 import HTTPProvider
from web3._utils.request import make_post_request


class BatchHTTPProvider(HTTPProvider):
    def make_batch_request(self, text):
        self.logger.debug(
            "Making request HTTP. URI: %s, Request: %s", self.endpoint_uri, text
        )
        request_data = text.encode("utf-8")
        raw_response = make_post_request(
            self.endpoint_uri, request_data, **self.get_request_kwargs()
        )
        response = self.decode_rpc_response(raw_response)
        self.logger.debug(
            "Getting response HTTP. URI: %s, " "Request: %s, Response: %s",
            self.endpoint_uri,
            text,
            response,
        )
        return response
